mod divider_row;
mod item_row;
mod message_row;
mod state_row;
mod verification_info_bar;

use self::divider_row::DividerRow;
use self::item_row::ItemRow;
use self::state_row::StateRow;
use self::verification_info_bar::VerificationInfoBar;

use adw::subclass::prelude::*;
use gtk::{
    gdk, glib, glib::clone, glib::signal::Inhibit, prelude::*, subclass::prelude::*,
    CompositeTemplate,
};
use matrix_sdk::ruma::events::room::message::{
    EmoteMessageEventContent, FormattedBody, MessageType, RoomMessageEventContent,
    TextMessageEventContent,
};
use sourceview::prelude::*;

use crate::components::{CustomEntry, Pill, RoomTitle};
use crate::session::content::{MarkdownPopover, RoomDetails};
use crate::session::room::{Item, Room, RoomType, Timeline};
use crate::session::user::UserExt;

mod imp {
    use super::*;
    use crate::Application;
    use glib::{signal::SignalHandlerId, subclass::InitializingObject};
    use std::cell::{Cell, RefCell};

    #[derive(Debug, Default, CompositeTemplate)]
    #[template(resource = "/org/gnome/FractalNext/content-room-history.ui")]
    pub struct RoomHistory {
        pub compact: Cell<bool>,
        pub room: RefCell<Option<Room>>,
        pub category_handler: RefCell<Option<SignalHandlerId>>,
        pub empty_timeline_handler: RefCell<Option<SignalHandlerId>>,
        pub loading_timeline_handler: RefCell<Option<SignalHandlerId>>,
        pub md_enabled: Cell<bool>,
        pub is_auto_scrolling: Cell<bool>,
        pub sticky: Cell<bool>,
        #[template_child]
        pub headerbar: TemplateChild<adw::HeaderBar>,
        #[template_child]
        pub room_title: TemplateChild<RoomTitle>,
        #[template_child]
        pub room_menu: TemplateChild<gtk::MenuButton>,
        #[template_child]
        pub listview: TemplateChild<gtk::ListView>,
        #[template_child]
        pub content: TemplateChild<gtk::Widget>,
        #[template_child]
        pub scrolled_window: TemplateChild<gtk::ScrolledWindow>,
        #[template_child]
        pub scroll_btn: TemplateChild<gtk::Button>,
        #[template_child]
        pub scroll_btn_revealer: TemplateChild<gtk::Revealer>,
        #[template_child]
        pub message_entry: TemplateChild<sourceview::View>,
        #[template_child]
        pub markdown_button: TemplateChild<gtk::MenuButton>,
        #[template_child]
        pub loading: TemplateChild<gtk::Spinner>,
        #[template_child]
        pub stack: TemplateChild<gtk::Stack>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for RoomHistory {
        const NAME: &'static str = "ContentRoomHistory";
        type Type = super::RoomHistory;
        type ParentType = adw::Bin;

        fn class_init(klass: &mut Self::Class) {
            CustomEntry::static_type();
            ItemRow::static_type();
            MarkdownPopover::static_type();
            VerificationInfoBar::static_type();
            Timeline::static_type();
            Self::bind_template(klass);
            klass.set_accessible_role(gtk::AccessibleRole::Group);
            klass.install_action(
                "room-history.send-text-message",
                None,
                move |widget, _, _| {
                    widget.send_text_message();
                },
            );
            klass.install_action("room-history.leave", None, move |widget, _, _| {
                widget.leave();
            });

            klass.install_action("room-history.details", None, move |widget, _, _| {
                widget.open_room_details("general");
            });
            klass.install_action("room-history.invite-members", None, move |widget, _, _| {
                widget.open_invite_members();
            });

            klass.install_action("room-history.scroll-down", None, move |widget, _, _| {
                widget.scroll_down();
            });
        }

        fn instance_init(obj: &InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for RoomHistory {
        fn properties() -> &'static [glib::ParamSpec] {
            use once_cell::sync::Lazy;
            static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
                vec![
                    glib::ParamSpecBoolean::new(
                        "compact",
                        "Compact",
                        "Whether a compact view is used",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    glib::ParamSpecObject::new(
                        "room",
                        "Room",
                        "The room currently shown",
                        Room::static_type(),
                        glib::ParamFlags::READWRITE | glib::ParamFlags::EXPLICIT_NOTIFY,
                    ),
                    glib::ParamSpecBoolean::new(
                        "empty",
                        "Empty",
                        "Whether there is currently a room shown",
                        false,
                        glib::ParamFlags::READABLE,
                    ),
                    glib::ParamSpecBoolean::new(
                        "markdown-enabled",
                        "Markdown enabled",
                        "Whether outgoing messages should be interpreted as markdown",
                        false,
                        glib::ParamFlags::READWRITE,
                    ),
                    glib::ParamSpecBoolean::new(
                        "sticky",
                        "Sticky",
                        "Whether the room history should stick to the newest message in the timeline",
                        true,
                        glib::ParamFlags::READWRITE,
                    ),
                ]
            });

            PROPERTIES.as_ref()
        }

        fn set_property(
            &self,
            obj: &Self::Type,
            _id: usize,
            value: &glib::Value,
            pspec: &glib::ParamSpec,
        ) {
            match pspec.name() {
                "compact" => {
                    let compact = value.get().unwrap();
                    self.compact.set(compact);
                }
                "room" => {
                    let room = value.get().unwrap();
                    obj.set_room(room);
                }
                "markdown-enabled" => {
                    let md_enabled = value.get().unwrap();
                    self.md_enabled.set(md_enabled);
                    self.markdown_button.set_icon_name(if md_enabled {
                        "format-indent-more-symbolic"
                    } else {
                        "format-justify-left-symbolic"
                    });
                }
                "sticky" => obj.set_sticky(value.get().unwrap()),
                _ => unimplemented!(),
            }
        }

        fn property(&self, obj: &Self::Type, _id: usize, pspec: &glib::ParamSpec) -> glib::Value {
            match pspec.name() {
                "compact" => self.compact.get().to_value(),
                "room" => obj.room().to_value(),
                "empty" => obj.room().is_none().to_value(),
                "markdown-enabled" => self.md_enabled.get().to_value(),
                "sticky" => obj.sticky().to_value(),
                _ => unimplemented!(),
            }
        }

        fn constructed(&self, obj: &Self::Type) {
            // Needed to use the natural height of GtkPictures
            self.listview
                .set_vscroll_policy(gtk::ScrollablePolicy::Natural);

            self.listview
                .connect_activate(clone!(@weak obj => move |listview, pos| {
                    if let Some(item) = listview
                        .model()
                        .and_then(|model| model.item(pos))
                        .and_then(|o| o.downcast::<Item>().ok())
                    {
                        if let Some(event) = item.event() {
                            if let Some(room) = obj.room() {
                                room.session().show_media(event);
                            }
                        }
                    }
                }));

            obj.set_sticky(true);
            let adj = self.listview.vadjustment().unwrap();

            adj.connect_value_changed(clone!(@weak obj => move |adj| {
                let priv_ = imp::RoomHistory::from_instance(&obj);

                if priv_.is_auto_scrolling.get() {
                    if adj.value() + adj.page_size() == adj.upper() {
                        priv_.is_auto_scrolling.set(false);
                        obj.set_sticky(true);
                    }
                } else {
                    obj.set_sticky(adj.value() + adj.page_size() == adj.upper());
                }
                obj.load_more_messages(adj);
            }));
            adj.connect_upper_notify(clone!(@weak obj => move |adj| {
                if obj.sticky() {
                    obj.scroll_down();
                }
                obj.load_more_messages(adj);
            }));

            let key_events = gtk::EventControllerKey::new();
            self.message_entry.add_controller(&key_events);

            key_events
                .connect_key_pressed(clone!(@weak obj => @default-return Inhibit(false), move |_, key, _, modifier| {
                if !modifier.contains(gdk::ModifierType::SHIFT_MASK) && (key == gdk::Key::Return || key == gdk::Key::KP_Enter) {
                    obj.activate_action("room-history.send-text-message", None).unwrap();
                    Inhibit(true)
                } else {
                    Inhibit(false)
                }
            }));

            let buffer = self
                .message_entry
                .buffer()
                .downcast::<sourceview::Buffer>()
                .unwrap();

            buffer.connect_text_notify(clone!(@weak obj => move |buffer| {
               let (start_iter, end_iter) = buffer.bounds();
               obj.action_set_enabled("room-history.send-text-message", start_iter != end_iter);
            }));
            crate::utils::setup_style_scheme(&buffer);

            let (start_iter, end_iter) = buffer.bounds();
            obj.action_set_enabled("room-history.send-text-message", start_iter != end_iter);

            let md_lang = sourceview::LanguageManager::default().language("markdown");
            buffer.set_language(md_lang.as_ref());
            obj.bind_property("markdown-enabled", &buffer, "highlight-syntax")
                .flags(glib::BindingFlags::SYNC_CREATE)
                .build();

            let settings = Application::default().settings();
            settings
                .bind("markdown-enabled", obj, "markdown-enabled")
                .build();

            self.parent_constructed(obj);
        }
    }

    impl WidgetImpl for RoomHistory {}
    impl BinImpl for RoomHistory {}
}

glib::wrapper! {
    pub struct RoomHistory(ObjectSubclass<imp::RoomHistory>)
        @extends gtk::Widget, adw::Bin, @implements gtk::Accessible;
}

impl RoomHistory {
    pub fn new() -> Self {
        glib::Object::new(&[]).expect("Failed to create RoomHistory")
    }

    pub fn set_room(&self, room: Option<Room>) {
        let priv_ = imp::RoomHistory::from_instance(self);

        if self.room() == room {
            return;
        }

        if let Some(category_handler) = priv_.category_handler.take() {
            if let Some(room) = self.room() {
                room.disconnect(category_handler);
            }
        }

        if let Some(empty_timeline_handler) = priv_.empty_timeline_handler.take() {
            if let Some(room) = self.room() {
                room.timeline().disconnect(empty_timeline_handler);
            }
        }

        if let Some(loading_timeline_handler) = priv_.loading_timeline_handler.take() {
            if let Some(room) = self.room() {
                room.timeline().disconnect(loading_timeline_handler);
            }
        }

        if let Some(ref room) = room {
            let handler_id = room.connect_notify_local(
                Some("category"),
                clone!(@weak self as obj => move |_, _| {
                        obj.update_room_state();
                }),
            );

            priv_.category_handler.replace(Some(handler_id));

            let handler_id = room.timeline().connect_notify_local(
                Some("empty"),
                clone!(@weak self as obj => move |_, _| {
                        obj.set_empty_timeline();
                }),
            );

            priv_.empty_timeline_handler.replace(Some(handler_id));

            let handler_id = room.timeline().connect_notify_local(
                Some("loading"),
                clone!(@weak self as obj => move |timeline, _| {
                    // We need to make sure that we loaded enough events to fill the `ScrolledWindow`
                    let priv_ = imp::RoomHistory::from_instance(&obj);
                    if !timeline.loading() {
                        let adj = priv_.listview.vadjustment().unwrap();
                        obj.load_more_messages(&adj);
                    }
                }),
            );

            priv_.loading_timeline_handler.replace(Some(handler_id));

            room.load_members();
        }

        // TODO: use gtk::MultiSelection to allow selection
        let model = room
            .as_ref()
            .map(|room| gtk::NoSelection::new(Some(room.timeline())));

        priv_.listview.set_model(model.as_ref());
        priv_.room.replace(room);
        let adj = priv_.listview.vadjustment().unwrap();
        self.load_more_messages(&adj);
        self.update_room_state();
        self.set_empty_timeline();
        self.notify("room");
        self.notify("empty");
    }

    pub fn room(&self) -> Option<Room> {
        let priv_ = imp::RoomHistory::from_instance(self);
        priv_.room.borrow().clone()
    }

    pub fn send_text_message(&self) {
        let priv_ = imp::RoomHistory::from_instance(self);
        let buffer = priv_.message_entry.buffer();
        let (start_iter, end_iter) = buffer.bounds();
        let body_len = buffer.text(&start_iter, &end_iter, true).len();

        let is_markdown = priv_.md_enabled.get();
        let mut has_mentions = false;
        let mut plain_body = String::with_capacity(body_len);
        // formatted_body is Markdown if is_markdown is true, and HTML if false.
        let mut formatted_body = String::with_capacity(body_len);
        // uncopied_text_location is the start of the text we haven't copied to
        // plain_body and formatted_body.
        let mut uncopied_text_location = start_iter;

        let mut iter = start_iter;
        loop {
            if let Some(anchor) = iter.child_anchor() {
                let widgets = anchor.widgets();
                let pill = widgets.first().unwrap().downcast_ref::<Pill>().unwrap();
                let (url, label) = pill
                    .user()
                    .map(|user| {
                        (
                            user.user_id().matrix_to_url().to_string(),
                            user.display_name(),
                        )
                    })
                    .or_else(|| {
                        pill.room().map(|room| {
                            (
                                // No server name needed. matrix.to URIs for mentions aren't routable
                                room.room_id().matrix_to_url([]).to_string(),
                                room.display_name(),
                            )
                        })
                    })
                    .unwrap();

                // Add more uncopied characters from message
                let some_text = buffer.text(&uncopied_text_location, &iter, false);
                plain_body.push_str(&some_text);
                formatted_body.push_str(&some_text);
                uncopied_text_location = iter;

                // Add mention
                has_mentions = true;
                plain_body.push_str(&label);
                formatted_body.push_str(&if is_markdown {
                    format!("[{}]({})", label, url)
                } else {
                    format!("<a href='{}'>{}</a>", url, label)
                });
            }
            if !iter.forward_char() {
                // Add remaining uncopied characters
                let some_text = buffer.text(&uncopied_text_location, &iter, false);
                plain_body.push_str(&some_text);
                formatted_body.push_str(&some_text);
                break;
            }
        }

        let is_emote = plain_body.starts_with("/me ");
        if is_emote {
            plain_body.replace_range(.."/me ".len(), "");
            formatted_body.replace_range(.."/me ".len(), "");
        }

        let html_body = if is_markdown {
            FormattedBody::markdown(formatted_body).map(|b| b.body)
        } else if has_mentions {
            // Already formatted with HTML
            Some(formatted_body)
        } else {
            None
        };

        let content = RoomMessageEventContent::new(if is_emote {
            MessageType::Emote(if let Some(html_body) = html_body {
                EmoteMessageEventContent::html(plain_body, html_body)
            } else {
                EmoteMessageEventContent::plain(plain_body)
            })
        } else {
            MessageType::Text(if let Some(html_body) = html_body {
                TextMessageEventContent::html(plain_body, html_body)
            } else {
                TextMessageEventContent::plain(plain_body)
            })
        });

        self.room().unwrap().send_message(content);
        buffer.set_text("");
    }

    pub fn leave(&self) {
        let priv_ = imp::RoomHistory::from_instance(self);

        if let Some(room) = &*priv_.room.borrow() {
            room.set_category(RoomType::Left);
        }
    }

    /// Opens the room details on the page with the given name.
    pub fn open_room_details(&self, page_name: &str) {
        if let Some(room) = self.room() {
            let window = RoomDetails::new(&self.parent_window(), &room);
            window.set_property("visible-page-name", page_name);
            window.show();
        }
    }

    pub fn open_invite_members(&self) {
        if let Some(room) = self.room() {
            let window = RoomDetails::new(&self.parent_window(), &room);
            window.set_property("visible-page-name", "members");
            window.present_invite_subpage();
            window.show();
        }
    }

    fn update_room_state(&self) {
        let priv_ = imp::RoomHistory::from_instance(self);

        if let Some(room) = &*priv_.room.borrow() {
            if room.category() == RoomType::Left {
                self.action_set_enabled("room-history.leave", false);
                priv_.room_menu.hide();
            } else {
                self.action_set_enabled("room-history.leave", true);
                priv_.room_menu.show();
            }
        }
    }

    fn set_empty_timeline(&self) {
        let priv_ = imp::RoomHistory::from_instance(self);

        if let Some(room) = &*priv_.room.borrow() {
            if room.timeline().is_empty() {
                priv_.stack.set_visible_child(&*priv_.loading);
            } else {
                priv_.stack.set_visible_child(&*priv_.content);
            }
        }
    }

    fn load_more_messages(&self, adj: &gtk::Adjustment) {
        // Load more messages when the user gets close to the end of the known room history
        // Use the page size twice to detect if the user gets close to the end
        if let Some(room) = self.room() {
            if adj.value() < adj.page_size() * 2.0
                || adj.upper() <= adj.page_size() / 2.0
                || room.timeline().is_empty()
            {
                room.timeline().load_previous_events();
            }
        }
    }

    /// Returns the parent GtkWindow containing this widget.
    fn parent_window(&self) -> Option<gtk::Window> {
        self.root()?.downcast().ok()
    }

    pub fn sticky(&self) -> bool {
        let priv_ = imp::RoomHistory::from_instance(self);

        priv_.sticky.get()
    }

    pub fn set_sticky(&self, sticky: bool) {
        let priv_ = imp::RoomHistory::from_instance(self);

        if self.sticky() == sticky {
            return;
        }

        priv_.scroll_btn_revealer.set_reveal_child(!sticky);

        priv_.sticky.set(sticky);
        self.notify("sticky");
    }

    /// Scroll to the newest message in the timeline
    pub fn scroll_down(&self) {
        let priv_ = imp::RoomHistory::from_instance(self);

        priv_.is_auto_scrolling.set(true);

        priv_
            .scrolled_window
            .emit_by_name::<bool>("scroll-child", &[&gtk::ScrollType::End, &false]);
    }
}

impl Default for RoomHistory {
    fn default() -> Self {
        Self::new()
    }
}
